"""
Created on 22 Apr 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.data.ingestion.duplicate_publication import DuplicatePublication


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('PubMonAPI/DuplicatePublication',
                   'https://nfema83f18.execute-api.us-west-2.amazonaws.com/default/DuplicatePublication')


# --------------------------------------------------------------------------------------------------------------------

class DuplicatePublicationManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find_summaries(self, token, summary_class):
        url = '/'.join([Endpoint.url(), 'summaries'])

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return sorted([summary_class.construct_from_jdict(jdict) for jdict in response.json()])


    def find_for_device(self, token, device):
        url = '/'.join([Endpoint.url(), 'device'])
        params = {'device': device}

        response = requests.get(url, headers=self._token_headers(token), params=params)
        self._check_response(response)

        return sorted([DuplicatePublication.construct_from_jdict(jdict) for jdict in response.json()])


    def delete_for_device(self, token, device):
        url = '/'.join([Endpoint.url(), 'device'])
        params = {'device': device}

        response = requests.delete(url, headers=self._token_headers(token), params=params)
        self._check_response(response)



"""
Created on 8 Aug 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

separated from client to remove dependency on requests package
"""
import json
from collections import OrderedDict
from urllib.parse import parse_qs, urlparse

from scs_core.aws.client.api_intercourse import APIResponse
from scs_core.aws.client_traffic.client_traffic import ClientTrafficLocus, ClientTrafficReport

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class ClientTrafficRequest(ClientTrafficLocus):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        endpoint = qsp.get('endpoint')
        clients = json.loads(qsp.get('clients'))
        period = qsp.get('period')

        aggregate = json.loads(qsp.get('aggregate'))

        return cls(endpoint, clients, period, aggregate)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        endpoint = jdict.get('endpoint')
        clients = jdict.get('clients')
        period = jdict.get('period')

        aggregate = jdict.get('aggregate')

        return cls(endpoint, clients, period, aggregate)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, endpoint, clients, period, aggregate):
        """
        Constructor
        """
        super().__init__(endpoint, clients, period)

        self.__aggregate = bool(aggregate)                          # bool


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {
            'endpoint': self.endpoint,
            'clients': json.dumps(self.client),
            'period': self.period,
            'aggregate': json.dumps(self.aggregate)
        }

        return params


    def as_json(self, **kwargs):
        jdict = OrderedDict()

        if self.endpoint is not None:
            jdict['endpoint'] = self.endpoint

        jdict['clients'] = self.client
        jdict['period'] = self.period

        jdict['aggregate'] = self.aggregate

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def aggregate(self):
        return self.__aggregate


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ClientTrafficRequest:{endpoint:%s, clients:%s, period:%s, aggregate:%s}" % \
            (self.endpoint, self.client, self.period, self.aggregate)


# --------------------------------------------------------------------------------------------------------------------

class ClientTrafficResponse(APIResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        items = []
        if jdict.get('Items'):
            for item_jdict in jdict.get('Items'):
                item = ClientTrafficReport.construct_from_jdict(item_jdict)
                items.append(item)

        next_request = jdict.get('next')

        return cls(items, next_request=next_request)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, items, next_request=None):
        """
        Constructor
        """
        self.__items = items                                # list of ClientTrafficReport
        self.__next_request = next_request                  # dict (lambda-to-lambda) or URL string (API gateway)


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, _):
        return parse_qs(urlparse(self.next_request).query)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, **kwargs):
        jdict = OrderedDict()

        if self.items is not None:
            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_request is not None:
            jdict['next'] = self.next_request

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def items(self):
        return self.__items


    @property
    def next_request(self):
        return self.__next_request


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ClientTrafficResponse:{items:%s, next_request:%s}" %  (Str.collection(self.items), self.next_request)



"""
Created on 9 Apr 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"topic": "ricardo/gatwick/loc/1/gases", "device": "scs-bgx-507", "rec": "2019-05-10T09:17:39Z",
"expiry": "2019-05-17T09:17:46Z"}
"""

from collections import OrderedDict

from scs_core.aws.client.api_intercourse import APIResponse
from scs_core.data.datetime import LocalizedDatetime


# --------------------------------------------------------------------------------------------------------------------

class TopicOrigin(APIResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        topic = jdict.get('topic')
        device = jdict.get('device')
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        expiry = LocalizedDatetime.construct_from_iso8601(jdict.get('expiry'))

        return cls(topic, device, rec, expiry)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, device, rec, expiry):
        """
        Constructor
        """
        self.__topic = topic                            # string
        self.__device = device                          # string
        self.__rec = rec                                # LocalizedDatetime
        self.__expiry = expiry                          # LocalizedDatetime or None


    def __lt__(self, other):
        if self.rec < other.rec:
            return True

        if self.rec > other.rec:
            return False

        if self.topic < other.topic:
            return True

        if self.topic > other.topic:
            return True

        return self.device < other.device


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, params):
        raise NotImplementedError


    @property
    def next_request(self):
        raise NotImplementedError


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, **kwargs):
        jdict = OrderedDict()

        jdict['topic'] = self.topic
        jdict['device'] = self.device
        jdict['rec'] = self.rec
        jdict['expiry'] = self.expiry

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def device(self):
        return self.__device


    @property
    def rec(self):
        return self.__rec


    @property
    def expiry(self):
        return self.__expiry


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{topic:%s, device:%s, rec:%s, expiry:%s}" % \
               (self.topic, self.device, self.rec, self.expiry)



"""
Created on 9 Apr 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.manager.topic_origin.topic_origin import TopicOrigin


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('OriginAPI/TopicOrigin',
                   'https://7qos9nwc7k.execute-api.us-west-2.amazonaws.com/default/TopicOrigin')


# --------------------------------------------------------------------------------------------------------------------

class TopicOriginFinder(APIClient):
    """
    classdocs
    """

    BLOCK_SIZE = 100                    # maximum number of topics per request

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find_for_topics(self, token, topics):
        if self._reporter:
            self._reporter.reset()

        origins = []

        for index in range(0, len(topics), self.BLOCK_SIZE):
            request = topics[index:index + self.BLOCK_SIZE]
            params = {"topics": json.dumps(request)}

            response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=params)
            self._check_response(response)

            block = [TopicOrigin.construct_from_jdict(jdict) for jdict in response.json()]
            origins.extend(block)

            if self._reporter:
                self._reporter.print(len(block))

        return sorted(origins)



"""
Created on 9 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"model": "ICP", "altitude": 100}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class PressureConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "pressure_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None) if skeleton else None

        model = jdict.get('model')
        altitude = jdict.get('altitude')

        return cls(model, altitude)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, model, altitude):
        """
        Constructor
        """
        super().__init__()

        self.__model = model                    # string
        self.__altitude = altitude              # int, 'GPS' or None


    def __eq__(self, other):
        try:
            return self.model == other.model or self.altitude == other.altitude

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def model(self):
        return self.__model


    @property
    def altitude(self):
        return self.__altitude


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, **kwargs):
        jdict = OrderedDict()

        jdict['model'] = self.__model
        jdict['altitude'] = self.__altitude

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PressureConf(core):{model:%s, altitude:%s}" % (self.model, self.altitude)



"""
Created on 7 June 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A catalogue entry for a machine learning model, for a specific PM size and model

document example:
<see scs_core.aws.model.pmx.compendia>
"""

import os

from collections import OrderedDict

from scs_core.data.json import JSONable, JSONCatalogueEntry
from scs_core.data.lin_regress import LinRegress
from scs_core.data.path_dict import PathDict
from scs_core.data.str import Str

from scs_core.model.catalogue.term import PrimaryTerm, SecondaryTerm, Term
from scs_core.model.catalogue.training_period import TrainingPeriod

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ReferenceCompendium(JSONable):
    """
    classdocs
    """

    REFERENCE_ENVIRONMENT = [
        'temp',
        'rh',
        'pres'
    ]

    @classmethod
    def is_target(cls, prefix, path, target):
        return cls.__is_term(prefix, path, [target.lower()])


    @classmethod
    def is_environment(cls, prefix, path):
        return cls.__is_term(prefix, path, cls.REFERENCE_ENVIRONMENT)


    @classmethod
    def __is_term(cls, prefix, path, term_set):
        if not path.startswith(prefix):
            return False

        return Term.suffix(prefix, path).lower() in term_set


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, prefix, target_path, reference_targets, reference_environments):
        target = Term.construct(target_path, reference_targets, prec=1)

        environment = {Term.suffix(prefix, path): Term.construct(path, values, prec=3)
                       for path, values in reference_environments.items()}

        return cls(target, environment)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        target = jdict.get('target')

        environment = {path: SecondaryTerm.construct_from_jdict(term_jdict)
                       for path, term_jdict in jdict.get('environment').items()}

        return cls(target, environment)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, target, environment):
        """
        Constructor
        """
        self.__target = target                          # Term
        self.__environment = environment                # dict of path: Term


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def target(self):
        return self.__target


    @property
    def environment(self):
        return self.__environment


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, **kwargs):
        jdict = OrderedDict()

        jdict['target'] = self.target
        jdict['environment'] = self.environment

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ReferenceCompendium:{target:%s, environment:%s}" % (self.target, self.environment)


# --------------------------------------------------------------------------------------------------------------------

class ModelCompendium(JSONCatalogueEntry):
    """
    classdocs
    """

    DEVICE_PRIMARIES = [
        'pmx.val.bin:0',
        'pmx.val.bin:1',
        'pmx.val.bin:2',
        'pmx.val.bin:3',
        'pmx.val.bin:4',
        'pmx.val.bin:5',
        'pmx.val.bin:6',
        'pmx.val.bin:7',
        'pmx.val.bin:8',
        'pmx.val.bin:9',
        'pmx.val.bin:10',
        'pmx.val.bin:11',
        'pmx.val.bin:12',
        'pmx.val.bin:13',
        'pmx.val.bin:14',
        'pmx.val.bin:15',
        'pmx.val.bin:16',
        'pmx.val.bin:17',
        'pmx.val.bin:18',
        'pmx.val.bin:19',
        'pmx.val.bin:20',
        'pmx.val.bin:21',
        'pmx.val.bin:22',
        'pmx.val.bin:23'
    ]

    DEVICE_VB_SECONDARIES = [
        'meteo.val.hmd.cur',
        'meteo.val.hmd.slp15min',
        'meteo.val.tmp.cur',
        'meteo.val.tmp.slp15min',
        'pmx.val.mtf1',
        'pmx.val.mtf3',
        'pmx.val.mtf5',
        'pmx.val.mtf7',
        'pmx.val.sfr',
        'pmx.val.sht.hmd.cur',
        'pmx.val.sht.hmd.slp15min',
        'pmx.val.sht.tmp.cur',
        'pmx.val.sht.tmp.slp15min'
    ]

    DEVICE_VE_SECONDARIES = [
        'meteo.exg.val.hmd.cur',
        'meteo.exg.val.hmd.slp15min',
        'meteo.val.tmp.cur',
        'meteo.val.tmp.slp15min',
        'pmx.val.mtf1',
        'pmx.val.mtf3',
        'pmx.val.mtf5',
        'pmx.val.mtf7',
        'pmx.val.sfr',
        'pmx.val.sht.hmd.cur',
        'pmx.val.sht.hmd.slp15min',
        'pmx.val.sht.tmp.cur',
        'pmx.val.sht.tmp.slp15min'
    ]

    DEVICE_SECONDARIES = {
        'vB': DEVICE_VB_SECONDARIES,
        'vE': DEVICE_VE_SECONDARIES
    }

    @classmethod
    def is_primary(cls, prefix, path):
        return cls.__is_term(prefix, path, cls.DEVICE_PRIMARIES)


    @classmethod
    def is_secondary(cls, prefix, path, model_interface):
        return cls.__is_term(prefix, path, cls.DEVICE_SECONDARIES[model_interface])


    @classmethod
    def __is_term(cls, prefix, path, term_set):
        if not path.startswith(prefix):
            return False

        return Term.suffix(prefix, path) in term_set


    # ----------------------------------------------------------------------------------------------------------------

    __CATALOGUE_NAME = 'compendia'

    @classmethod
    def catalogue_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), cls.__CATALOGUE_NAME)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, is_error_model, data_set, recs, device_prefix, primary_values, secondary_values, output_path,
                  outputs, reference_prefix, target_path, reference_targets, reference_environments):
        period = TrainingPeriod.construct(recs)

        primaries = {Term.suffix(device_prefix, path): PrimaryTerm.construct(path, values, prec=3)
                     for path, values in primary_values.items()}

        secondaries = {Term.suffix(device_prefix, path): SecondaryTerm.construct(path, values, prec=3)
                       for path, values in secondary_values.items()}

        reference = ReferenceCompendium.construct(reference_prefix, target_path, reference_targets,
                                                  reference_environments)

        output = Term.construct(output_path, outputs, prec=1)

        performance = LinRegress.construct(reference_targets, outputs, prec=3)

        return cls(is_error_model, data_set, period, primaries, secondaries, reference, output, performance)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        is_error_model = jdict.get('is-error-model', False)
        data_set = jdict.get('data-set')
        period = TrainingPeriod.construct_from_jdict(jdict.get('period'))

        primaries = {path: PrimaryTerm.construct_from_jdict(term_jdict)
                     for path, term_jdict in jdict.get('primaries').items()}

        secondaries = {path: SecondaryTerm.construct_from_jdict(term_jdict)
                       for path, term_jdict in jdict.get('secondaries').items()}

        reference = Term.construct_from_jdict(jdict.get('reference'))

        output = Term.construct_from_jdict(jdict.get('output'))

        performance = LinRegress.construct_from_jdict(jdict.get('performance'))

        return cls(is_error_model, data_set, period, primaries, secondaries, reference, output, performance)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, is_error_model, data_set, period, primaries, secondaries, reference, output, performance):
        """
        Constructor
        """
        self.__is_error_model = bool(is_error_model)        # bool
        self.__data_set = data_set                          # string
        self.__period = period                              # TrainingPeriod

        self.__primaries = primaries                        # dict of path: PrimaryTerm
        self.__secondaries = secondaries                    # dict of path: SecondaryTerm
        self.__reference = reference                        # ReferenceCompendium
        self.__output = output                              # Term

        self.__performance = performance                    # LinRegress

        self.__logger = Logging.getLogger()


    def __len__(self):
        return len(self.performance)


    # ----------------------------------------------------------------------------------------------------------------

    def is_in_bounds(self, datum: PathDict):
        for datum_path in datum.paths():
            if datum_path in self.secondaries:
                if not self.secondaries[datum_path].is_in_bounds(datum.node(datum_path)):
                    return False

        return True


    def postprocess(self, model_output):
        # self.__logger.info("postprocess - primary_path: %s vcal_excess: %s model_output: %s" %
        #                    (primary_path, vcal_excess, model_output))

        corrected_exg = (model_output - self.performance.intercept) / self.performance.slope

        return corrected_exg


    # ----------------------------------------------------------------------------------------------------------------

    def primary_paths(self):
        return [primary.path for primary in self.primaries.values()]


    def primary_term(self, path):
        try:
            return self.primaries[path]
        except KeyError:
            return None


    def secondary_paths(self):
        return [secondary.path for secondary in self.secondaries.values()]


    def secondary_term(self, path):
        try:
            return self.secondaries[path]
        except KeyError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        pieces = self.output.path.split('.')
        return '.'.join(pieces[1:])


    @property
    def species_name(self):
        pieces = self.output.path.split('.')
        return pieces[1]


    @property
    def model_name(self):
        pieces = self.output.path.split('.')
        return pieces[2]


    @property
    def device_name(self):
        pieces = self.output.path.split('.')
        return pieces[3]


    @property
    def period_name(self):
        pieces = self.output.path.split('.')
        return pieces[4]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def is_error_model(self):
        return self.__is_error_model


    @property
    def data_set(self):
        return self.__data_set


    @property
    def period(self):
        return self.__period


    @property
    def primaries(self):
        return self.__primaries


    @property
    def secondaries(self):
        return self.__secondaries


    @property
    def reference(self):
        return self.__reference


    @property
    def output(self):
        return self.__output


    @property
    def performance(self):
        return self.__performance


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, **kwargs):
        jdict = OrderedDict()

        jdict['is-error-model'] = self.is_error_model
        jdict['data-set'] = self.data_set
        jdict['period'] = self.period

        jdict['primaries'] = self.primaries
        jdict['secondaries'] = self.secondaries
        jdict['reference'] = self.reference
        jdict['output'] = self.output

        jdict['performance'] = self.performance

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        primaries = Str.collection(self.primaries)
        secondaries = Str.collection(self.secondaries)

        return "ModelCompendium:{name:%s, is_error_model:%s, data_set:%s, period:%s, primaries:%s, secondaries:%s, " \
               "reference:%s, output:%s, performance:%s}" %  \
               (self.name, self.is_error_model, self.data_set, self.period, primaries, secondaries,
                self.reference, self.output, self.performance)



"""
Created on 19 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

separated from client to remove dependency on requests package
"""

from abc import ABC, abstractmethod
from http import HTTPStatus

from scs_core.aws.data.http_response import HTTPResponse
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class APIResponse(JSONable, ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # server...

    def as_http(self, status=HTTPStatus.OK):
        response = HTTPResponse(status, self)

        return response.as_http()


    # ----------------------------------------------------------------------------------------------------------------
    # client...

    def start(self):
        return None


    @abstractmethod
    def next_params(self, params):
        pass


    @property
    @abstractmethod
    def next_request(self):
        pass


    @property
    def interval(self):
        return None
